//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2015 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// 1st Dec 2016: Ideal coherent turbo coded OFDM for massive MIMO.
// "ldiv" is not used for generating the impulse response of the
// constituent convolutional codes. Multiply the received signal with the
// Hermitian of the channel gain matrix.
// Code tested on SCILAB ver 6.0.0.
//******************************************************************************
//clear;
//getd(pwd());                    // Get all functions in the current directory.
timer();
//******************************************************************************
// Add a word of caution.
//******************************************************************************
mprintf('Running this code will append to the data files\n');
mprintf('u_ber.txt and c_ber.txt.\n');
mprintf('Is this ok?.\n');
mprintf('\n');
pause;
//******************************************************************************
// Initialize the random number generators with the date.
//******************************************************************************
n=getdate("s");
grand("setsd",n);
//******************************************************************************
// Constants.
//******************************************************************************
SNR_DB=4.25;                                               // SNR per bit in dB.
MAX_FRAMES=100000;                      // No. of OFDM symbols to be simulated.
OFDM_SIZE=512;                                         // Length of OFDM symbol.
DATA_SIZE=OFDM_SIZE;
NUM_TX=512;                                        // No. of transmit antennas.
NUM_RX=NUM_TX;
NUM_RE_TX=2;                                          // No. of retransmissions.
//******************************************************************************
x=modulo(2*OFDM_SIZE,NUM_TX);
if x ~= 0 then
 mprintf('OFDM_SIZE/NUM_TX must be an integer.\n');
end
NUM_BLOCKS=2*OFDM_SIZE/NUM_TX;
//******************************************************************************
// Open files for appending.
//******************************************************************************
format(20);
x=string(SNR_DB);
y=strsubst(x, '.', '_');
z1=string(NUM_TX);
z2=string(NUM_RE_TX);
z3=string(MAX_FRAMES);
fd_u_ber_w=mopen(pwd()+'/u_ber_'+'tx'+z1+'_retx'+z2+'_snr'+y+...
                '_fr'+z3+'_tc16.txt','at');
fd_c_ber_w=mopen(pwd()+'/c_ber_'+'tx'+z1+'_retx'+z2+'_snr'+y+...
                '_fr'+z3+'_tc16.txt','at');
clear x;
clear y;
//******************************************************************************
// SNR parameters.
//******************************************************************************
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=2.0*NUM_RX*NUM_RE_TX/(10^(0.1*SNR_DB));        // Var per dim of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// QPSK symbols. Real part (data) maps to MSB. Imag. part (parity) maps to
// LSB.
//******************************************************************************
CONSTELL_SIZE=4;
QPSK_Map(1)=  1+%i;
QPSK_Map(2)=  1-%i;
QPSK_Map(3)= -1+%i;
QPSK_Map(4)= -1-%i;
for i=1:CONSTELL_SIZE
 QPSK_SYM(i,:)=QPSK_Map(i)*ones(1,2*DATA_SIZE);
end
//******************************************************************************
// Generator matrix for the rate-1/2 systematic convolutional encoder:
// [1 (1+D^2)/(1+D+D^2)]. Refer to Digital Communications and
// Signal Processing by K Vasudevan, Universities Press, for the notation
// followed.
//******************************************************************************
NC=[1 0 1 1 1];                                       // Numerator coefficients.
DC=[1 1 0 0 1];                                     // Denominator coefficients.
NC_Size=size(NC);
NUM_INPUTS=2;                                      // No. of input combinations.
NUM_STATES=2^(NC_Size(2)-1);                           // No. of trellis states.
RSC_MEM=NC_Size(2)-1;                                 // No. of memory elements.
[N_State,P_State,P_Ip,Op,Ga_Inx,Gb_Inx]=Get_Trellis();
RSC_Gen_Vec=Get_RSC_Gen_Vec();                          // Get impulse response.
//******************************************************************************
// Get the interleaver and deinterleaver map for the turbo code.
//******************************************************************************
[Iv_Map,Div_Map]=Get_Iv_Div_Map(DATA_SIZE);
//******************************************************************************
// Minimum exponent.
//******************************************************************************
MIN_EXPO= -30.0;
//******************************************************************************
// Initialize gammas.
//******************************************************************************
My_Gamma1=zeros(CONSTELL_SIZE,DATA_SIZE);
My_Gamma2=zeros(CONSTELL_SIZE,DATA_SIZE);
//******************************************************************************
Ber=0;                                          // Uncoded bit-error-rate (BER).
C_Ber=0;                                            // BER after turbo decoding.
Est_Noise_Var=(8.0*NUM_TX*(NUM_TX-1.0)*FADE_VAR^2+...
               4.0*NUM_TX*FADE_VAR*NOISE_VAR)/NUM_RE_TX;
//******************************************************************************
// Simulation.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
//******************************************************************************
// Generate coded QPSK symbols and append postamble for 1st and 2nd
// transmitter.
//******************************************************************************
 [C1,C2,My_Data]=Get_Coded_QPSK(DATA_SIZE,RSC_Gen_Vec,Div_Map);
 C=[C1 C2];
//******************************************************************************
// Iterate over the number of retransmissions.
//******************************************************************************
 Rx_Sig=zeros(1,2*DATA_SIZE);
 Gain_Vec=zeros(1,2*DATA_SIZE);
 for Re_Tx=1:NUM_RE_TX
//******************************************************************************
// Iterate over the number of blocks. Symbols in one block are transmitted
// simultaneously over NUM_TX transmit antennas.
//******************************************************************************
  for Block_Cnt=1:NUM_BLOCKS
   Block_Start_Inx=(Block_Cnt-1)*NUM_TX+1;
   Block_End_Inx=Block_Cnt*NUM_TX;
   Block_Sym=C(Block_Start_Inx:Block_End_Inx);
//******************************************************************************
// Get channel.
//******************************************************************************
   Chan=Get_MIMO_Flat_Channel();
//  [U,S,V]=svd(Chan);
//  Chan=U;
   Block_Gain_Vec=diag(Chan'*Chan);
//******************************************************************************
// Get channel output.
//******************************************************************************
   Chan_Op=Get_MIMO_Channel_Op();
//******************************************************************************
// Multiply the received signal with the conjugate transpose of the channel
// matrix.
//******************************************************************************
   Chan_Op=Chan'*Chan_Op;
//   Chan_Op=inv(Chan)*Chan_Op;
   Rx_Sig(Block_Start_Inx:Block_End_Inx)=...
   Rx_Sig(Block_Start_Inx:Block_End_Inx)+Chan_Op.';
   Gain_Vec(Block_Start_Inx:Block_End_Inx)=...
   Gain_Vec(Block_Start_Inx:Block_End_Inx)+Block_Gain_Vec.';
  end
 end
// pause;
//******************************************************************************
// Compute squared Euclidean distances.
//******************************************************************************
 for tx_sym=1:CONSTELL_SIZE
  Dist(tx_sym,:)=abs(Rx_Sig-QPSK_SYM(tx_sym,:).*real(Gain_Vec)).^2;
//  Dist(tx_sym,:)=abs(Rx_Sig-QPSK_SYM(tx_sym,:)).^2;                     // ZF.
 end
//******************************************************************************
  [M1,K1]=min(Dist(1,:),Dist(2,:),Dist(3,:),Dist(4,:));            // Min. dist.
  Rx_Sym=QPSK_Map(K1);                                         // Column vector.
  Ber=Ber+nnz(real(C1-Rx_Sym(1:DATA_SIZE).'));                   // Uncoded BER.
//  pause;
//******************************************************************************
// Estimate the 2D noise variance.
//******************************************************************************
//  Est_Noise_Var=real((C-Rx_Sig)*(C-Rx_Sig)')/(2.0*DATA_SIZE);           // ZF.
//  pause;
//******************************************************************************
// Compute gammas.
//******************************************************************************
  My_Exp1= -Dist(:,1:DATA_SIZE)/Est_Noise_Var;
  My_Exp2= -Dist(:,DATA_SIZE+1:2*DATA_SIZE)/Est_Noise_Var;
  My_Exp1=My_Exp1-repmat(max(My_Exp1,'r'),CONSTELL_SIZE,1);
  My_Exp2=My_Exp2-repmat(max(My_Exp2,'r'),CONSTELL_SIZE,1);
  Exp1_Inx=find(My_Exp1 < MIN_EXPO);
  Exp2_Inx=find(My_Exp2 < MIN_EXPO);
  My_Exp1(Exp1_Inx)= MIN_EXPO;
  My_Exp2(Exp2_Inx)= MIN_EXPO;
  My_Gamma1(:,:)=exp(My_Exp1);
  My_Gamma2(:,:)=exp(My_Exp2);
//  pause;
//******************************************************************************
// The MAP (BCJR) algorithm.
//******************************************************************************
  Prob=0.5*ones(2,DATA_SIZE);
//******************************************************************************
//  [Prob]=BCJR(Prob);
//******************************************************************************
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process_End(Prob);
//******************************************************************************
  [M6,K6]=max(Prob(1,:),Prob(2,:));                   // Max. prob. and indices.
  Rx_Sym1=1-2*(K6-1);
  C_Ber=C_Ber+nnz(My_Data-Rx_Sym1);
//******************************************************************************
end
//******************************************************************************
F_U_Ber=Ber/(MAX_FRAMES*DATA_SIZE);                   // Fractional uncoded BER.
F_C_Ber=C_Ber/(MAX_FRAMES*DATA_SIZE);                    //Fractional coded BER.
mfprintf(fd_u_ber_w,'%f %.20f\n',SNR_DB,F_U_Ber);
mfprintf(fd_c_ber_w,'%f %.20f\n',SNR_DB,F_C_Ber);
mclose(fd_u_ber_w);
mclose(fd_c_ber_w);
timer()
//******************************************************************************
